__short_name__ = 'aaa'
__iam__ = 'flake8-{}'.format(__short_name__)
__version__ = '0.11.0'

__author__ = 'James Cooke'
__copyright__ = '2018 - 2020, {}'.format(__author__)

__description__ = 'A Flake8 plugin that checks Python tests follow the Arrange-Act-Assert pattern'
__email__ = 'github@jamescooke.info'
